import pydoc
from typing import Dict, Any
import pytest

from pydoctest.parsers.sphinx_parser import SphinxParser
from pydoctest.exceptions import ParseException

import tests.test_parsers.sphinx_class


class TestSphinxParser():

    def test_parse_exception_get_parameters(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.IncorrectTestClass.func_parse_exception)
        with pytest.raises(ParseException) as exc_info:
            parser.get_parameters(doc, tests.test_parsers.sphinx_class)

    def test_empty_func(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.empty_func)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 0, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == type(None), f"SphinxParser failed assertion"

    def test_func_returns_none(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_returns_none)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 1, f"SphinxParser failed assertion"
        assert arguments[0].type == int, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == type(None), f"SphinxParser failed assertion"

    def test_func_returns_int(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_returns_int)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 0, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == int, f"SphinxParser failed assertion"

    def test_func_has_arg_returns_arg(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_has_arg_returns_arg)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 1, f"SphinxParser failed assertion"
        assert arguments[0].type == int, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == float, f"SphinxParser failed assertion"

    def test_func_has_raises_doc(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_has_raises_doc)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 1, f"SphinxParser failed assertion"
        assert arguments[0].type == int, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == int, f"SphinxParser failed assertion"

    def test_func_with_multiline_summary(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_with_multiline_summary)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 1, f"SphinxParser failed assertion"
        assert arguments[0].type == int, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == int, f"SphinxParser failed assertion"

    def test_get_summary_multiline_summary(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_with_multiline_summary)

        summary = parser.get_summary(doc, tests.test_parsers.sphinx_class)
        assert summary is not None
        assert summary == """Lorem ipsum dolor sit amet, consectetur adipiscing elit.
Mauris tellus diam, iaculis et nisl sit amet, tristique sodales massa.
Proin lacinia faucibus ex a scelerisque. Fusce mauris orci, finibus a cursus vitae, luctus sit amet erat.
Ut dignissim elit nec nisi faucibus lobortis.
Proin varius mi lectus, at gravida nisl dapibus et. Nunc ac sagittis sapien.
Praesent tincidunt ac tellus ut mollis. Integer tincidunt pretium diam, quis aliquet turpis volutpat quis.
Nulla at est facilisis, scelerisque ipsum in, interdum dolor. Vivamus fermentum placerat mattis.
Ut nec augue nec ex sodales ornare vel sit amet urna. Nunc scelerisque risus nisi, quis pharetra nibh fermentum auctor.
Donec ultrices lectus eu mauris lacinia, nec tincidunt tortor facilisis.
Sed condimentum elit non metus sagittis tempor. Cras mollis lacus lacus, vitae placerat quam laoreet id.""", f"SphinxParser failed assertion"

    def test_get_summary_empty_summary(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_no_summary)
        arguments = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(arguments) == 0, f"SphinxParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == type(None), f"SphinxParser failed assertion"

        summary = parser.get_summary(doc, tests.test_parsers.sphinx_class)
        assert summary is None, f"SphinxParser failed assertion"

    def test_func_with_raise_and_args_and_return(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.RaisesClass.func_with_raise_and_args_and_return)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_raise_and_args(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.RaisesClass.func_with_raise_and_args)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

        parameters = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(parameters) == 2
        assert parameters[0].name == 'a'
        assert parameters[0].type == int

        assert parameters[1].name == 'b'
        assert parameters[1].type == float

    def test_func_with_raise(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.RaisesClass.func_with_raise)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_generics(self) -> None:
        parser = SphinxParser()
        doc = pydoc.getdoc(tests.test_parsers.sphinx_class.CorrectTestClass.func_with_generics)
        parameters = parser.get_parameters(doc, tests.test_parsers.sphinx_class)
        assert len(parameters) == 1
        assert parameters[0].type == Dict[str, Any]
        assert parameters[0].name == 'a_a'

        return_type = parser.get_return_type(doc, tests.test_parsers.sphinx_class)
        assert return_type == Dict[str, Any]



from typing import Optional
from pydoctest.configuration import Verbosity
from pydoctest.reporters.reporter import Reporter
from pydoctest.validation import ClassValidationResult, FunctionValidationResult, ModuleValidationResult, ResultType, ValidationResult


SUCCESS = "OK"
FAILED = "FAIL"
SKIPPED = "SKIPPED"


class TextReporter(Reporter):
    def get_output(self, result: ValidationResult) -> str:
        """Returns the text output by walking the ValidationResult object.

        Args:
            result (ValidationResult): The results from running Pydoctest.

        Returns:
            str: The output to be returned.
        """
        output = ""
        for module_result in result.module_results:
            output += self.get_module_output(module_result)

        return output

    def get_module_output(self, result: ModuleValidationResult) -> str:
        """Returns the text output from the result object from walking the module.

        Args:
            result (ModuleValidationResult): The result from running Pydoctest on the module.

        Returns:
            str: The output of the module.
        """
        output = ""
        if result.fail_reason != "":
            output = f"{result.fail_reason}\n"

        for f_r in result.function_results:
            output += self.get_function_output(f_r)

        for c_r in result.class_results:
            output += self.get_class_output(c_r)

        return output

    def get_function_output(self, result: FunctionValidationResult, class_name: Optional[str] = None) -> str:
        """Returns the text output from the result object from the function.

        Args:
            result (FunctionValidationResult): The result from running Pydoctest on the function.
            class_name (Optional[str]): Optionally which class this function is run within.
        Returns:
            str: The output from the function.
        """
        module = ""
        if result.module.__file__:
            # Try to get just workspace relative path
            module = result.module.__file__.replace(self.config.working_directory, "")

        function_name = result.function.__name__
        class_name_if_exists = class_name + '::' if class_name is not None else ''
        if result.result == ResultType.OK:
            if self.config.verbosity == Verbosity.SHOW_ALL:
                return f"{module}::{class_name_if_exists}{function_name} {SUCCESS}\n"
            return ""

        if result.result == ResultType.FAILED:
            return f"{module}::{class_name_if_exists}{function_name} {FAILED} | {result.fail_reason}\n"

        if result.result == ResultType.NO_DOC and self.config.fail_on_missing_docstring:
            return f"{module}::{class_name_if_exists}{function_name} is missing a docstring\n"

        return ""

    def get_class_output(self, result: ClassValidationResult) -> str:
        """Returns the text output from the result object from the class.

        Args:
            result (ClassValidationResult): The result from running Pydoctest on the class.

        Returns:
            str: The output of the class and its functions.
        """
        output = ""
        for fn in result.function_results:
            output += self.get_function_output(fn, class_name=result.class_name)
        return output



class ExampleClass():

    # Range should be entire docstring (5..12)
    def func_no_summary(self, a: int) -> int:
        """

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass

    # Range should be signature (16)
    def func_no_docstring(self, a: int) -> int:
        pass

    # Range should be entire docstring (22..32)
    # TODO: optimize later with specific line perhaps)
    def func_parse_exception(self, a: int) -> int:
        """[summary]

        Args:
            THISDOESNTPARSE

        Raises:
            THISDOESNTPARSE

        Returns:
            THISDOESNTPARSE
        """
        pass

    # Range should be the line that contains the type (38..44)
    # TODO: Optimization, return only the line with return type
    def func_return_type_differ(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Returns:
            bool: [description]
        """
        pass

    # Range should be entire docstring (51..58)
    # TODO: Optimization: If too many args in docstring, highlight which one.
    def func_number_of_arguments_differ(self, a: int, b: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass

    # Range should be the line with the name in (64..71)
    # TODO: Optimization, highlight only 'b' or 'b (int)' - depends on format?
    def func_argument_name_differ(self, a: int) -> int:
        """[summary]

        Args:
            b (int): [description]

        Returns:
            int: [description]
        """
        pass

    # Range should be line with argument in it (77..84)
    # TODO: Optimization, highlight only 'b (int)'
    def func_argument_type_differ(self, a: int) -> int:
        """[summary]

        Args:
            a (bool): [description]

        Returns:
            int: [description]
        """
        pass

    # Range should be entire docstring (89..99)
    def func_number_of_raised_exceptions_differ(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Raises:
            RuntimeError: [description]
            ValueError: [description]

        Returns:
            int: [description]
        """
        if 2 == 3:
            raise RuntimeError()

        if 2 == 4:
            raise ValueError()

        if 2 == 5:
            raise IndexError()

    # Range should be entire docstring (113..123)
    # TODO: Optimization, highlight specific conflicts.
    def func_listed_exceptions_not_match(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Raises:
            ValueError: [description]

        Returns:
            int: [description]
        """
        raise RuntimeError()



import pydoc

from pydoctest.configuration import Configuration
from pydoctest.validation import validate_function, ResultType

import tests.test_ranges.example_class
import tests.test_class.raises_class


class TestRanges():
    def test_func_no_summary(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_missing_summary = True
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_no_summary, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == 'Function does not have a summary'
        assert result.range is not None
        assert result.range.start_line == 5
        assert result.range.end_line == 12

    def test_func_no_docstring(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_missing_docstring = True
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_no_docstring, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == 'Function does not have a docstring'
        assert result.range is not None
        assert result.range.start_line == 16
        assert result.range.end_line == 16

    def test_func_parse_exception(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_parse_exception, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Unable to parse docstring' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 22
        assert result.range.end_line == 32

    def test_func_return_type_differ(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_return_type_differ, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Return type differ' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 38
        assert result.range.end_line == 45

    def test_func_number_of_arguments_differ(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_number_of_arguments_differ, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Number of arguments differ' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 51
        assert result.range.end_line == 58

    def test_func_argument_name_differ(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_argument_name_differ, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Argument name differ' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 64
        assert result.range.end_line == 71

    def test_func_argument_type_differ(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_argument_type_differ, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Argument type differ' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 77
        assert result.range.end_line == 84

    def test_func_number_of_raised_exceptions_differ(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_number_of_raised_exceptions_differ, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Number of listed raised exceptions does not match actual' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 89
        assert result.range.end_line == 100

    def test_func_listed_exceptions_not_match(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_ranges.example_class.ExampleClass.func_listed_exceptions_not_match, config, tests.test_ranges.example_class)
        assert result.result == ResultType.FAILED
        assert 'Listed raised exceptions does not match actual' in result.fail_reason
        assert result.range is not None
        assert result.range.start_line == 113
        assert result.range.end_line == 123



import json
from pydoctest.reporters.reporter import Reporter
from pydoctest.validation import ValidationResult


class JSONReporter(Reporter):
    def get_output(self, result: ValidationResult) -> str:
        """Returns the JSON output by walking the ValidationResult object.

        Args:
            result (ValidationResult): The results from running Pydoctest

        Returns:
            str: The JSON output to be returned.
        """
        dict_result = result.to_dict()
        return json.dumps(dict_result)





def failing_global_function(a: int) -> int:
    """[summary]
                            <-- missing args
    Returns:
        int: [description]
    """
    pass


class IncorrectTestClass():
    def empty_func(self) -> None:
        """Empty func

        Args:
            a (int): [description]          <-- should have no args
        """
        pass

    def func_returns_none(self, a: int) -> None:
        """Func returns None

        Args:
            a (int): [description]

        Returns:
            int: [description]              <-- should not be there
        """
        pass

    def func_returns_int(self) -> int:
        """Func returns int

        Returns:
            bool: [description]             <-- incorrect
        """
        pass

    def func_has_arg_returns_arg(self, a: int) -> float:
        """Func takes argument and returns argument
                        <-- should have args and returns
        """
        pass

    def func_has_raises_doc(self, a: int) -> int:
        """[summary]

        Raises:
            Exception: [description]           <-- Only have raises
        """
        raise Exception()

    def func_with_multiline_summary(self, a: int) -> int:
        """Lorem ipsum dolor sit amet, consectetur adipiscing elit.
        Mauris tellus diam, iaculis et nisl sit amet, tristique sodales massa.
        Proin lacinia faucibus ex a scelerisque. Fusce mauris orci, finibus a cursus vitae, luctus sit amet erat.
        Ut dignissim elit nec nisi faucibus lobortis.
        Proin varius mi lectus, at gravida nisl dapibus et. Nunc ac sagittis sapien.
        Praesent tincidunt ac tellus ut mollis. Integer tincidunt pretium diam, quis aliquet turpis volutpat quis.
        Nulla at est facilisis, scelerisque ipsum in, interdum dolor. Vivamus fermentum placerat mattis.
        Ut nec augue nec ex sodales ornare vel sit amet urna. Nunc scelerisque risus nisi, quis pharetra nibh fermentum auctor.
        Donec ultrices lectus eu mauris lacinia, nec tincidunt tortor facilisis.
        Sed condimentum elit non metus sagittis tempor. Cras mollis lacus lacus, vitae placerat quam laoreet id.

        Args:
            a (int): [description]
            b (int): [description]          <-- incorrect

        Returns:
            int: [description]
        """
        pass

    def func_name_mismatch(self, a: int) -> int:
        """[summary]

        Args:
            b (int): [description]      <-- name is not 'a'

        Returns:
            int: [description]
        """
        pass

    def func_type_mismatch(self, a: int) -> int:
        """[summary]

        Args:
            a (float): [description]        <-- float is not int

        Returns:
            int: [description]
        """
        pass

    def func_parse_exception(self, a: int) -> int:
        """[summary]

        Args:
            THISDOESNTPARSE

        Raises:
            THISDOESNTPARSE

        Returns:
            THISDOESNTPARSE
        """
        pass



from enum import IntEnum
import inspect
from sys import modules
import types

from types import FunctionType, ModuleType
from typing import Any, Dict, List, Optional, Type

from pydoctest.logging import log
from pydoctest.configuration import Configuration
from pydoctest.parsers.parser import Parameter
from pydoctest.exceptions import ParseException
from pydoctest.utilities import get_exceptions_raised, is_excluded_function


class Range():

    def __init__(self, start_line: int, end_line: int, start_character: int, end_character: int) -> None:
        """Creates a new range object, used for indicating errors in vscode.

        TODO: Are these zero- or one-indexed?

        Args:
            start_line (int): Where the range starts.
            end_line (int): Where the range ends.
            start_character (int): The column the range starts in (in start_line).
            end_character (int): The column the range ends in (in end_line).
        """
        self.start_line: int = start_line
        self.end_line: int = end_line
        self.start_character: int = start_character
        self.end_character: int = end_character

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this class to dict, which is useful for the JSONReporter.

        Returns:
            Dict[str, Any]: The range.
        """
        return {
            'start_line': self.start_line,
            'end_line': self.end_line,
            'start_character': self.start_character,
            'end_character': self.end_character
        }


class ValidationCounts():
    def __init__(self) -> None:
        """Helper class for storing counts from running Pydoctest.
        """
        self.module_count = 0
        self.functions_succeeded = 0
        self.functions_failed = 0
        self.functions_skipped = 0

    def get_total(self) -> int:
        """Returns the total number of functions executed.

        Returns:
            int: The total count.
        """
        return self.functions_succeeded + self.functions_failed + self.functions_skipped


class ResultType(IntEnum):
    NOT_RUN = 0
    OK = 1
    FAILED = 2
    SKIPPED = 3
    NO_DOC = 4


class Result():
    def __init__(self) -> None:
        """Base Result class for storing result and fail_reason.
        """
        # TODO: Rename result -> Status?
        self.result: ResultType = ResultType.NOT_RUN
        self.fail_reason: str = ""

    # TODO: Rather than have all Result-subclasses implement this, we can probably make a generic function
    def to_dict(self) -> Dict[str, Any]:
        """Serializes this class to dict, which is useful for the JSONReporter.

        Returns:
            Dict[str, Any]: The result and fail-reason.
        """
        return { 'result': self.result, 'fail_reason': self.fail_reason }


class FunctionValidationResult(Result):
    def __init__(self, function: FunctionType, module: ModuleType) -> None:
        """Result class for storing results of testing functions.

        Args:
            function (FunctionType): A reference to the function that was tested.
            module (ModuleType): The module containing the function - used when outputting text results to identify the file.
        """
        super().__init__()
        self.module = module
        self.function = function
        self.range: Optional[Range] = None

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this class to dict, which is useful for the JSONReporter.

        Returns:
            Dict[str, Any]: The result, fail_reason and function_name.
        """
        return {
            **super().to_dict(),
            'function': str(self.function),
            'range': self.range.to_dict() if self.range else None
        }


class ClassValidationResult(Result):
    def __init__(self, class_name: str) -> None:
        """Result class for storing results of testing classes.

        Args:
            class_name (str): The name of the class.
        """
        super().__init__()
        self.class_name = class_name
        self.function_results: List[FunctionValidationResult] = []

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this class to dict, which is useful for the JSONReporter.

        Returns:
            Dict[str, Any]: The result, fail_reason, class_name and function_results.
        """
        return {
            **super().to_dict(),
            'class_name': self.class_name,
            'function_results': [
                r.to_dict() for r in self.function_results
            ]
        }


class ModuleValidationResult(Result):
    def __init__(self, module_path: str) -> None:
        """Result class for storing results of testing modules.

        Args:
            module_path (str): The path to the module tested.
        """
        super().__init__()
        self.module_path = module_path
        self.function_results: List[FunctionValidationResult] = []
        self.class_results: List[ClassValidationResult] = []

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this class to dict, which is useful for the JSONReporter.

        Returns:
            Dict[str, Any]: The result, fail_reason, module_path, function_results and class_results.
        """
        return {
            **super().to_dict(),
            'module_path': self.module_path,
            'function_results': [
                r.to_dict() for r in self.function_results
            ],
            'class_results': [
                r.to_dict() for r in self.class_results
            ]
        }


class ValidationResult(Result):
    def __init__(self) -> None:
        """Result class for storing results of running pydoctest on a project.
        """
        super().__init__()
        self.module_results: List[ModuleValidationResult] = []

    def to_dict(self) -> Dict[str, Any]:
        """Serializes this class to dict, which is useful for the JSONReporter.

        Returns:
            Dict[str, Any]: The result, fail_reason and module_results.
        """
        return {
            **super().to_dict(),
            'module_results': [
                r.to_dict() for r in self.module_results
            ]
        }

    def get_counts(self) -> ValidationCounts:
        """Counts the failed, succeeded and skipped tests of running pydoctest.

        Returns:
            ValidationCounts: The counts object.
        """
        counts = ValidationCounts()
        counts.module_count = len(self.module_results)

        def increment(fn_result: FunctionValidationResult) -> None:
            if fn_result.result == ResultType.FAILED:
                counts.functions_failed += 1
            elif fn_result.result == ResultType.OK:
                counts.functions_succeeded += 1
            else:
                counts.functions_skipped += 1

        for m in self.module_results:
            for fn in m.function_results:
                increment(fn)
            for c in m.class_results:
                for fn in c.function_results:
                    increment(fn)
        return counts


def __get_docstring_range(fn: FunctionType, module_type: ModuleType, docstring: Optional[str] = None) -> Optional[Range]:
    """Returns the range for the docstring.

    Args:
        fn (FunctionType): The function to validate.
        module_type (ModuleType): The module from which the function was extracted.
        docstring (Optional[str]): Optionally, the docstring.

    Returns:
        Optional[Range]: The range, if found.
    """
    lines, line_number = inspect.getsourcelines(fn)
    start_line, end_line = -1, -1
    for i, l in enumerate(lines):
        if l.count('"""') == 2:
            # One liner
            return Range(line_number + i, line_number + i, 0, 0)

        if '"""' in l:
            if start_line == -1:
                start_line = line_number + i
            elif end_line == -1:
                end_line = line_number + i
                return Range(start_line, end_line, 0, 0)
    return None


def validate_function(fn: FunctionType, config: Configuration, module_type: ModuleType) -> FunctionValidationResult:
    """Validates the docstring of a function against its signature.

    Args:
        fn (FunctionType): The function to validate.
        config (Configuration): The configuration to use while validating.
        module_type (ModuleType): The module from which the function was extracted.

    Returns:
        FunctionValidationResult: The result of validating this function.
    """
    log(f"Validating function: {fn}")
    result = FunctionValidationResult(fn, module_type)

    doc = inspect.getdoc(fn)
    if not doc:
        if config.fail_on_missing_docstring:
            result.result = ResultType.FAILED
            result.fail_reason = f"Function does not have a docstring"
            _, line_number = inspect.getsourcelines(fn)
            result.range = Range(line_number, line_number, 0, 0)
        else:
            result.result = ResultType.NO_DOC
        return result

    parser = config.get_parser()

    summary = parser.get_summary(doc, module_type)
    if not summary and config.fail_on_missing_summary:
        result.result = ResultType.FAILED
        result.fail_reason = f"Function does not have a summary"
        result.range = __get_docstring_range(fn, module_type, doc)
        return result

    sig = inspect.signature(fn)
    sig_parameters = [Parameter(name, proxy.annotation) for name, proxy in sig.parameters.items() if name != "self"]
    sig_return_type = type(None) if sig.return_annotation is None else sig.return_annotation

    try:
        doc_parameters = parser.get_parameters(doc, module_type)
        doc_return_type = parser.get_return_type(doc, module_type)
    except ParseException as e:
        result.result = ResultType.FAILED
        result.fail_reason = f"Unable to parse docstring: {str(e)}"
        result.range = __get_docstring_range(fn, module_type, doc)
        return result

    # Validate return type
    if sig_return_type != doc_return_type:
        result.result = ResultType.FAILED
        result.fail_reason = f"Return type differ. Expected (from signature) {sig_return_type}, but got (in docs) {doc_return_type}."
        result.range = __get_docstring_range(fn, module_type, doc)
        return result

    # Validate equal number of parameters
    if len(sig_parameters) != len(doc_parameters):
        result.result = ResultType.FAILED
        result.fail_reason = f"Number of arguments differ. Expected (from signature) {len(sig_parameters)} arguments, but found (in docs) {len(doc_parameters)}."
        result.range = __get_docstring_range(fn, module_type, doc)
        return result

    # Validate name and type of function parameters
    for sigparam, docparam in zip(sig_parameters, doc_parameters):
        if sigparam.name != docparam.name:
            result.result = ResultType.FAILED
            result.fail_reason = f"Argument name differ. Expected (from signature) '{sigparam.name}', but got (in docs) '{docparam.name}'"
            result.range = __get_docstring_range(fn, module_type, doc)
            return result

        # NOTE: Optional[str] == Union[str, None] # True
        if sigparam.type != docparam.type:
            result.result = ResultType.FAILED
            result.fail_reason = f"Argument type differ. Argument '{sigparam.name}' was expected (from signature) to have type '{sigparam.type}', but has (in docs) type '{docparam.type}'"
            result.range = __get_docstring_range(fn, module_type, doc)
            return result

    # Validate exceptions raised
    if config.fail_on_raises_section:
        try:
            sig_exceptions = get_exceptions_raised(fn, module_type)
            doc_exceptions = parser.get_exceptions_raised(doc)

            if len(sig_exceptions) != len(doc_exceptions):
                result.result = ResultType.FAILED
                result.fail_reason = f"Number of listed raised exceptions does not match actual. Doc: {doc_exceptions}, expected: {sig_exceptions}"
                result.range = __get_docstring_range(fn, module_type, doc)
                return result

            intersection = set(sig_exceptions) - set(doc_exceptions)
            if len(intersection) > 0:
                result.result = ResultType.FAILED
                result.fail_reason = f"Listed raised exceptions does not match actual. Docstring: {doc_exceptions}, expected: {sig_exceptions}"
                result.range = __get_docstring_range(fn, module_type, doc)
                return result
        except ParseException as e:
            result.result = ResultType.FAILED
            result.fail_reason = f"Unable to parse docstring: {str(e)}"
            result.range = __get_docstring_range(fn, module_type, doc)
            return result

    result.result = ResultType.OK
    return result


def validate_class(class_instance: Any, config: Configuration, module_type: ModuleType) -> ClassValidationResult:
    """Validates the class by validating each of its methods.

    Args:
        class_instance (Any): A class to validate.
        config (Configuration): The configuration to use while validating.
        module_type (ModuleType): The module from which the class was extracted.

    Returns:
        ClassValidationResult: The result of validating this class.
    """
    log(f"Validating class: {class_instance}")
    class_result = ClassValidationResult(class_instance.__name__)

    for name, item in inspect.getmembers(class_instance):
        if inspect.isfunction(item) and item.__module__ == module_type.__name__:
            if name not in class_instance.__dict__ or item != class_instance.__dict__[name]:
                continue

            # Check if method is excluded
            if is_excluded_function(name, config.exclude_methods):
                continue

            function_result = validate_function(item, config, module_type)
            if function_result.result == ResultType.FAILED:
                class_result.result = ResultType.FAILED

            class_result.function_results.append(function_result)

    # If result has not been changed at this point, it must be OK
    if class_result.result == ResultType.NOT_RUN:
        class_result.result = ResultType.OK

    return class_result



class RaisesClass():
    def func_with_raise(self) -> None:
        """
        Raises:
            RuntimeError: [description]
            ValueError: [description]
            IndexError: [description]
        """
        if 2 == 3:
            raise RuntimeError()

        if 2 == 4:
            raise ValueError()

        if 2 == 5:
            raise IndexError()

    def func_with_raise_and_args(self) -> None:
        """
        Raises:
            RuntimeError: [description]
            ValueError: [description]
            IndexError: [description]
        """
        if 2 == 3:
            raise RuntimeError()

        if 2 == 4:
            raise ValueError()

        if 2 == 5:
            raise IndexError()

    def func_with_raise_and_args_and_return(self, a: int, b: float) -> bool:
        """[summary]

        Args:
            a (int): [description]
            b (float): [description]

        Raises:
            RuntimeError: [description]
            ValueError: [description]
            IndexError: [description]

        Returns:
            bool: [description]
        """
        if 2 == 3:
            raise RuntimeError()

        if 2 == 4:
            raise ValueError()

        if 2 == 5:
            raise IndexError()

        return True

    def func_with_missing_raise(self) -> None:
        """
        """
        if 2 == 5:
            raise IndexError()

    def func_with_incorrect_raise(self) -> None:
        """[summary]

        Raises:
            RuntimeError: [description]
            IndexError: [description]
        """
        if 2 == 4:
            raise ValueError()

        if 2 == 5:
            raise IndexError()

    def func_with_raise_count_mismatch(self) -> None:
        """[summary]

        Raises:
            RuntimeError: [description]
            IndexError: [description]
        """
        if 2 == 3:
            raise RuntimeError()

        if 2 == 4:
            raise ValueError()

        if 2 == 5:
            raise IndexError()


from enum import Enum


class MyEnum(Enum):
    A = 0
    B = 1
    C = 2


class ExampleClass():
    def a(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass



import pydoc
from typing_extensions import Literal
import pytest

from pydoctest.parsers.google_parser import GoogleParser
from pydoctest.exceptions import ParseException

import tests.test_class.incorrect_class
import tests.test_parsers.google_class


class TestGoogleParser():
    def test_parse_exception_get_parameters(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.incorrect_class.IncorrectTestClass.func_parse_exception)
        with pytest.raises(ParseException) as exc_info:
            parser.get_parameters(doc, tests.test_class.incorrect_class)

    def test_parse_exception_get_return_type(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.incorrect_class.IncorrectTestClass.func_parse_exception)
        with pytest.raises(ParseException) as exc_info:
            parser.get_return_type(doc, tests.test_class.incorrect_class)

    def test_get_exceptions_raised(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.incorrect_class.IncorrectTestClass.func_parse_exception)
        with pytest.raises(ParseException) as exc_info:
            parser.get_exceptions_raised(doc)

    def test_empty_func(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.empty_func)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 0, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == type(None), f"GoogleParser failed assertion"

    def test_func_returns_none(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_returns_none)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 1, f"GoogleParser failed assertion"
        assert arguments[0].type == int, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == type(None), f"GoogleParser failed assertion"

    def test_func_returns_int(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_returns_int)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 0, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == int, f"GoogleParser failed assertion"

    def test_func_has_arg_returns_arg(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_has_arg_returns_arg)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 1, f"GoogleParser failed assertion"
        assert arguments[0].type == int, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == float, f"GoogleParser failed assertion"

    def test_func_has_raises_doc(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_has_raises_doc)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 1, f"GoogleParser failed assertion"
        assert arguments[0].type == int, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == int, f"GoogleParser failed assertion"

    def test_func_with_multiline_summary(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_with_multiline_summary)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 1, f"GoogleParser failed assertion"
        assert arguments[0].type == int, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == int, f"GoogleParser failed assertion"

    def test_get_summary_multiline_summary(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_with_multiline_summary)

        summary = parser.get_summary(doc, tests.test_class.correct_class)
        assert summary is not None
        assert len(summary) > 0, f"GoogleParser failed assertion"
        assert(len([x for x in summary if x == '\n']) > 1), f"GoogleParser failed assertion"

    def test_get_summary_empty_summary(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.correct_class.CorrectTestClass.func_no_summary)
        arguments = parser.get_parameters(doc, tests.test_class.correct_class)
        assert len(arguments) == 0, f"GoogleParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_class.correct_class)
        assert return_type == type(None), f"GoogleParser failed assertion"

        summary = parser.get_summary(doc, tests.test_class.correct_class)
        assert summary is None, f"GoogleParser failed assertion"

    def test_func_with_raise_and_args_and_return(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.raises_class.RaisesClass.func_with_raise_and_args_and_return)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_raise_and_args(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.raises_class.RaisesClass.func_with_raise_and_args)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_raise(self) -> None:
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_class.raises_class.RaisesClass.func_with_raise)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_colon_in(self) -> None:
        """
        Solves: https://github.com/jepperaskdk/pydoctest/issues/29
        """
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_parsers.google_class.GoogleClass.function_with_colon)
        params = parser.get_parameters(doc, tests.test_parsers.google_class)

        assert len(params) == 2
        assert params[0].name == "a"
        assert params[0].type == int
        assert params[1].name == "b"
        assert params[1].type == int

    def test_function_with_literal(self) -> None:
        """
        Solves: https://github.com/jepperaskdk/pydoctest/issues/41
        """
        parser = GoogleParser()
        doc = pydoc.getdoc(tests.test_parsers.google_class.GoogleClass.function_with_literal)
        params = parser.get_parameters(doc, tests.test_parsers.google_class)

        assert len(params) == 1
        assert params[0].name == 'a'
        assert params[0].type == Literal['b', 'c']



import setuptools
import sys
import os

from pydoctest.version import VERSION


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydoctest",
    version=VERSION,
    author="Jeppe Rask",
    author_email="jepperaskdk@gmail.com",
    description="Test if doctype types match signatures",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jepperaskdk/pydoctest",
    project_urls={
        "Bug Tracker": "https://github.com/jepperaskdk/pydoctest/issues",
    },
    package_dir={"": "."},
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'pydoctest=pydoctest.main:main'
        ]
    },
    license_files=('LICENSE',),
    python_requires=">=3.6",
)



class CorrectTestClass():
    def empty_func(self) -> None:
        """Empty func
        """
        pass

    def func_returns_none(self, a: int) -> None:
        """Func returns None

        Args:
            a (int): [description]
        """
        pass

    def func_returns_int(self) -> int:
        """Func returns int

        Returns:
            int: [description]
        """
        pass

    def func_has_arg_returns_arg(self, a: int) -> float:
        """Func takes argument and returns argument

        Args:
            a (int): [description]

        Returns:
            float: [description]
        """
        pass

    def func_has_raises_doc(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Raises:
            Exception: [description]

        Returns:
            int: [description]
        """
        raise Exception()

    def func_with_multiline_summary(self, a: int) -> int:
        """Lorem ipsum dolor sit amet, consectetur adipiscing elit.
        Mauris tellus diam, iaculis et nisl sit amet, tristique sodales massa.
        Proin lacinia faucibus ex a scelerisque. Fusce mauris orci, finibus a cursus vitae, luctus sit amet erat.
        Ut dignissim elit nec nisi faucibus lobortis.
        Proin varius mi lectus, at gravida nisl dapibus et. Nunc ac sagittis sapien.
        Praesent tincidunt ac tellus ut mollis. Integer tincidunt pretium diam, quis aliquet turpis volutpat quis.
        Nulla at est facilisis, scelerisque ipsum in, interdum dolor. Vivamus fermentum placerat mattis.
        Ut nec augue nec ex sodales ornare vel sit amet urna. Nunc scelerisque risus nisi, quis pharetra nibh fermentum auctor.
        Donec ultrices lectus eu mauris lacinia, nec tincidunt tortor facilisis.
        Sed condimentum elit non metus sagittis tempor. Cras mollis lacus lacus, vitae placerat quam laoreet id.

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass

    def func_no_summary(self) -> None:
        """
        """
        pass
